@@ -1,26 +1,32 @@
 import base64
+import getpass
 import json
 import logging
 import os
 import os.path as op
-import requests
-import slugid
+import re
 import sys
 import tempfile
 import time
 import typing
+from pathlib import Path
+
+import requests
+from dotenv import load_dotenv
 
 import higlass.client as hgc
 import higlass.utils as hgu
+import slugid
 from higlass import Track
 
 # from higlass.utils import fill_filetype_and_datatype
 from resgen import aws
 
 # import resgen.utils as rgu
+logging.basicConfig(level=logging.INFO)
 logger = logging.getLogger(__name__)
 
-__version__ = "0.4.1"
+__version__ = "0.5.0"
 
 RESGEN_HOST = "https://resgen.io"
 RESGEN_BUCKET = "resgen"
@@ -34,6 +40,18 @@ RESGEN_AUTH0_DOMAIN = "https://auth.resgen.io"
 MAX_LIMIT = int(1e6)
 
 
+def parse_ucsc(hub_string):
+    # print("hub_string:", hub_string)
+
+    things = []
+    for hub_section in re.split("\n\n+", hub_string.strip()):
+        parts = [d.split(maxsplit=1) for d in hub_section.split("\n") if d[0] != "#"]
+        if parts:
+            things += [dict(parts)]
+
+    return things
+
+
 def update_tags(current_tags, **kwargs):
     """Update a list of tags to use for searching.
 
@@ -94,11 +112,16 @@ class ChromosomeInfo:
         return self.cum_chrom_lengths[chrom] + pos
 
     def to_abs_range(
-        self, chrom: str, start: int, end: int, padding: float = 0
+        self,
+        chrom: str,
+        start: int,
+        end: int,
+        padding: float = 0,
+        padding_abs: float = 0,
     ) -> typing.Tuple[int, int]:
         """Return a range along a chromosome with optional padding.
         """
-        padding_abs = (end - start) * padding
+        padding_abs += (end - start) * padding
 
         return [
             self.cum_chrom_lengths[chrom] + start - padding_abs,
@@ -412,7 +435,9 @@ class ResgenConnection:
 
         return json.loads(ret.content)
 
-    def upload_to_resgen_aws(self, filepath: str, prefix: str = None) -> str:
+    def upload_to_resgen_aws(
+        self, filepath: str, prefix: str = None, index_filepath=None
+    ) -> str:
         """
         Upload file to a resgen aws bucket.
 
@@ -435,23 +460,35 @@ class ResgenConnection:
 
         content = json.loads(ret.content)
         filename = op.split(filepath)[1]
+
         directory_path = f"{content['fileDirectory']}/{filename}"
         object_name = f"{content['uploadBucketPrefix']}/{filename}"
 
         logger.info("Uploading to aws object: %s", object_name)
+        index_directory_path = None
+        if index_filepath:
+            index_filename = op.split(index_filepath)[1]
+            index_object_name = f"{content['uploadBucketPrefix']}/{index_filename}"
+            index_directory_path = f"{content['fileDirectory']}/{index_filename}"
+            logger.info("Uploading to aws index object: %s", index_object_name)
 
         bucket = self.bucket
         if aws.upload_file(filepath, bucket, content, object_name):
-            return directory_path
+            if index_filepath:
+                if not aws.upload_file(
+                    index_filepath, bucket, content, index_object_name
+                ):
+                    return None
+            return (directory_path, index_directory_path)
 
-        return None
+        return (None, None)
 
     def update_dataset(
         self, uuid: str, metadata: typing.Dict[str, typing.Any]
     ) -> ResgenDataset:
         """Update the properties of a dataset."""
         new_metadata = {}
-        updatable_properties = ["name", "datafile", "tags"]
+        updatable_properties = ["name", "datafile", "tags", "description"]
 
         for key in metadata:
             if key not in updatable_properties:
@@ -461,6 +498,8 @@ class ResgenConnection:
                 )
         if "name" in metadata:
             new_metadata["name"] = metadata["name"]
+        if "description" in metadata:
+            new_metadata["description"] = metadata["description"]
         if "datafile" in metadata:
             new_metadata["datafile"] = metadata["datafile"]
         if "tags" in metadata:
@@ -487,6 +526,7 @@ class ResgenProject:
             conn: The resgen connection that this project was created with
             name: The name of the project. Not strictly necessary, but helpful
                 when stringifying this object.
+            try: Dry run... don't actually sync datasets
 
         """
         self.uuid = uuid
@@ -502,66 +542,129 @@ class ResgenProject:
 
         # raise NotImplementedError()
 
-    def add_dataset(self, filepath: str, download: bool = False):
-        """Add a dataset
+    def add_link_dataset(self, filepath: str, index_filepath: str = None):
+        """Add a dataset by downloading it from a remote source
 
         Args:
             filepath: The filename of the dataset to add. Can also be a url.
-            download: If the filepath is a url, download it and save it to our
-                datastore. Useful for files on ftp servers or servers that do
-                not have range requests.
+            index_filepath: The filename of the index for this dataset
 
         Returns:
             The uuid of the newly created dataset.
 
         """
-        if download:
-            ret = self.conn.authenticated_request(
-                requests.post,
-                f"{self.conn.host}/api/v1/tilesets/",
-                json={
-                    "datafile": filepath,
-                    "private": True,
-                    "project": self.uuid,
-                    "description": f"Downloaded from {filepath}",
-                    "download": True,
-                    "tags": [],
-                },
+        logger.info("Adding link dataset: %s", filepath)
+        body = {
+            "datafile": filepath,
+            "private": True,
+            "project": self.uuid,
+            "download": False,
+            "tags": [],
+        }
+
+        if index_filepath:
+            body["indexfile"] = index_filepath
+
+        ret = self.conn.authenticated_request(
+            requests.post, f"{self.conn.host}/api/v1/tilesets/", json=body,
+        )
+        content = json.loads(ret.content)
+        print("content", content)
+        return content["uuid"]
+
+    def add_download_dataset(self, filepath: str, index_filepath: str = None):
+        """Add a dataset by downloading it from a remote source
+
+        Args:
+            filepath: The filename of the dataset to add. Can also be a url.
+            index_filepath: The filename of the index for this dataset
+
+        Returns:
+            The uuid of the newly created dataset.
+
+        """
+        logger.info("Adding download dataset: %s", filepath)
+        body = {
+            "datafile": filepath,
+            "private": True,
+            "project": self.uuid,
+            "description": f"Downloaded from {filepath}",
+            "download": True,
+            "tags": [],
+        }
+
+        if index_filepath:
+            body["indexfile"] = index_filepath
+
+        ret = self.conn.authenticated_request(
+            requests.post, f"{self.conn.host}/api/v1/tilesets/", json=body,
+        )
+
+        if ret.status_code != 201:
+            logger.error(
+                "Error adding download dataset. Error code: %d, Content: %s",
+                ret.status_code,
+                ret.content,
             )
+            return
 
-            content = json.loads(ret.content)
+        content = json.loads(ret.content)
 
-            progress = {"downloaded": 0, "uploaded": 0, "filesize": 1}
-            while (
-                progress["downloaded"] < progress["filesize"]
-                or progress["uploaded"] < progress["filesize"]
-                or progress["downloaded"] == 0
-            ):
-                try:
-                    progress = self.conn.download_progress(content["uuid"])
-                except UnknownConnectionException:
-                    pass
-                time.sleep(0.5)
+        progress = {"downloaded": 0, "uploaded": 0, "filesize": 1}
+        while (
+            progress["downloaded"] < progress["filesize"]
+            or progress["uploaded"] < progress["filesize"]
+            or progress["downloaded"] == 0
+        ):
+            try:
+                progress = self.conn.download_progress(content["uuid"])
+            except UnknownConnectionException:
+                pass
+            time.sleep(0.5)
 
-                if progress["filesize"] > 0:
-                    percent_done = (
-                        100
-                        * (progress["downloaded"] + progress["uploaded"])
-                        / (2 * progress["filesize"])
-                    )
+            if progress["filesize"] > 0:
+                transferred = progress["downloaded"] + progress["uploaded"]
+                to_transfer = 2 * progress["filesize"]
+                percent_done = 100 * transferred / to_transfer
 
-                    sys.stdout.write(f"\r {percent_done:.2f}% Complete")
+                sys.stdout.write(
+                    f"\r {percent_done:.3f}% Complete ({transferred} of {to_transfer})"
+                )
 
-            return content["uuid"]
+        return content["uuid"]
 
-        directory_path = self.conn.upload_to_resgen_aws(filepath)
+    def add_upload_dataset(self, filepath: str, index_filepath: str = None):
+        """Add a dataset by uploading it to resgen
+
+        Args:
+            filepath: The filename of the dataset to add. Can also be a url.
+            index_filepath: The filename of the index for this dataset
+
+        Returns:
+            The uuid of the newly created dataset.
+
+        """
+        logger.info("Adding upload dataset: %s", filepath)
+
+        (directory_path, index_directory_path) = self.conn.upload_to_resgen_aws(
+            filepath, index_filepath=index_filepath
+        )
+
+        if not directory_path:
+            raise Exception("Error uploading to AWS")
 
         logger.info("Adding tileset entry for uploaded file: %s", directory_path)
 
+        body = {
+            "filepath": directory_path,
+            "project": self.uuid,
+        }
+
+        if index_directory_path:
+            body["indexpath"] = index_directory_path
+
         ret = self.conn.authenticated_request(
-            requests.post,
-            f"{self.conn.host}/api/v1/finish_file_upload/",
-            json={"filepath": directory_path, "project": self.uuid},
+            requests.post, f"{self.conn.host}/api/v1/finish_file_upload/", json=body,
         )
 
         if ret.status_code != 200:
@@ -570,6 +673,21 @@ class ResgenProject:
         content = json.loads(ret.content)
         return content["uuid"]
 
+    def add_dataset(
+        self,
+        filepath: str,
+        download: bool = False,
+        index_filepath: str = None,
+        sync_remote: bool = False,
+    ):
+        if download:
+            if sync_remote:
+                return self.add_download_dataset(filepath, index_filepath)
+            else:
+                return self.add_link_dataset(filepath, index_filepath)
+        else:
+            return self.add_upload_dataset(filepath, index_filepath)
+
     def delete_dataset(self, uuid: str):
         """Delete a dataset."""
         ret = self.conn.authenticated_request(
@@ -643,12 +761,79 @@ class ResgenProject:
         """String representation."""
         return f"{self.uuid[:8]}: {self.name}"
 
+    def sync_track_hub(self, base_url, dry=False):
+        """Sync a UCSC track hub."""
+        hub_url = f"{base_url}/hub.txt"
+
+        ret = requests.get(hub_url)
+
+        content = ret.content.decode("utf8")
+        hub_info = parse_ucsc(content)[0]
+
+        genomes_url = f'{base_url}/{hub_info["genomesFile"]}'
+        print("genomes_url:", genomes_url)
+
+        ret = requests.get(genomes_url)
+
+        content = ret.content.decode("utf8")
+        genome_infos = parse_ucsc(content)
+
+        for genome_info in genome_infos:
+            self.sync_genome(base_url, genome_info, dry)
+
+    def sync_genome(self, base_url, genome_info, dry=False):
+        """Sync a genome within a track hub."""
+        track_db_url = f"{base_url}/{genome_info['trackDb']}"
+        # print("track_db_url:", track_db_url)
+        ret = requests.get(track_db_url)
+        content = ret.content.decode("utf8")
+        genome_info_path = op.split(genome_info["trackDb"])[0]
+
+        track_infos = parse_ucsc(content)
+        for track in track_infos:
+            # print("------------------")
+            # print("track:", track)
+            # print("track:", track.keys())
+            type_parts = track["type"].split()
+            track_type = type_parts[0]
+
+            if track_type in ["bigWig", "bigBed"] and track.get("bigDataUrl"):
+                big_data_path = (
+                    f"{base_url}/{genome_info_path}/{track.get('bigDataUrl')}"
+                )
+
+                if track_type == "bigWig":
+                    datatype = "vector"
+                elif track_type == "bigBed" and type_parts[1] == 12:
+                    datatype = "gene-annotations"
+                else:
+                    datatype = "bedlike"
+
+                assembly = genome_info["genome"]
+                name = track.get("shortLabel")
+                description = track.get("longLabel")
+
+                if not dry:
+                    self.sync_dataset(
+                        big_data_path,
+                        sync_remote=False,
+                        datatype=datatype,
+                        filetype=track_type.lower(),
+                        assembly=assembly,
+                        name=name,
+                        description=description,
+                    )
+                else:
+                    print(f"Syncing: {big_data_path} assembly: {assembly}")
+
     def sync_dataset(
         self,
         filepath: str,
+        sync_remote: bool = False,
         filetype=None,
         datatype=None,
         assembly=None,
+        index_filepath=None,
         force_update: bool = False,
         **metadata,
     ):
@@ -663,7 +848,9 @@ class ResgenProject:
         If more than one dataset with this name exists, raise a ValueError.
 
         Args:
+
         """
+        logger.info("Syncing dataset: %s", filepath)
         if (
             filepath.startswith("http://")
             or filepath.startswith("https://")
@@ -689,12 +876,23 @@ class ResgenProject:
             raise ValueError("More than one matching dataset")
 
         if not matching_datasets:
-            uuid = self.add_dataset(filepath, download=download)
+            uuid = self.add_dataset(
+                filepath,
+                download=download,
+                index_filepath=index_filepath,
+                sync_remote=sync_remote,
+            )
         else:
+            logger.info("Found dataset with the same filepath, updating metadata")
             uuid = matching_datasets[0].data["uuid"]
 
             if force_update:
-                new_uuid = self.add_dataset(filepath, download=download)
+                new_uuid = self.add_dataset(
+                    filepath,
+                    download=download,
+                    index_filepath=index_filepath,
+                    sync_remote=sync_remote,
+                )
                 self.delete_dataset(uuid)
                 uuid = new_uuid
 
@@ -737,6 +935,17 @@ def connect(
     bucket: str = RESGEN_BUCKET,
 ) -> ResgenConnection:
     """Open a connection to resgen."""
+    env_path = Path.home() / ".resgen" / "credentials"
+
+    if env_path.exists():
+        load_dotenv(env_path)
+
+        username = os.getenv("RESGEN_USERNAME")
+        password = os.getenv("RESGEN_PASSWORD")
+    else:
+        username = input("Username:")
+        password = getpass.getpass("Password:")
+
     if username is None:
         username = os.getenv("RESGEN_USERNAME")
     if password is None:
